import java.io.*;
import java.net.*;

/**
 * Captures the clock accuracy, latency and clock skew between two servers.
 * To run:
 *   on server machine enter: java CheckLatency Server <port number>
 *   on client machine enter: java CheckLatency <hostname> <port number> <iterations>
 * Always start server machine first
 * When client completes test both client and server will stop.
 *
 * <hostname> can be a named host or its IP address.
 *
 */
public class CheckLatency {

    private static long clockResolution; // accuracy of clock

    /**
     * Detects resolution of the clock by calling System.getNanos() multiple times.
     * Multiple calls ensures that we obtain sample to greatest precision.
     * @return resolution in nanoseconds
     */
    private static long  testClockResolution() {
        final int numTrials = 50;
        int clockResolution [] = new int[9];

        System.out.println("Calculate clock accuracy with "+numTrials+ " samples.");

        for (int i=0; i<numTrials; i++) {
            long resolution=1000000000;
            long nanoTime = System.nanoTime();
            for (int j=0; j<9; j++) {
                if (nanoTime%resolution!=0) {
                    clockResolution[8-j]++;
                }
                resolution=resolution/10;
            }
        }

        long accuracy = 1;
        for (int i=0;i<9;i++) {
            if (clockResolution[i]!=0) {
                break;
            }
            accuracy=accuracy*10;
        }

        System.out.println("Cloud accuracy is "+accuracy+ " nanosecs.");
        return accuracy;
    }

    /**
     * Captures the latency and clock skew for two servers adopting accuracy of client clock
     * Should extend to use least accurate clock.
     *
     * @param hostName - hostname or ip address to establish connection
     * @param portNumber - port number
     * @param iterations - number of times to send request
     */
    private static void client(String hostName, int portNumber, int iterations) {
        System.out.println("Attempting to open socket connection to "+hostName+" on port "+portNumber+".");
        try (
                Socket timeSocket = new Socket(hostName, portNumber);
                PrintWriter out =
                        new PrintWriter(timeSocket.getOutputStream(), true);
                BufferedReader in =
                        new BufferedReader(
                                new InputStreamReader(timeSocket.getInputStream()))
        ) {
            long clientSendNanos; // sent timestamp
            long serverNanos; // time returned by server
            long clientReceiveNanos; // received timestamp
            String serverTimestamp; // string version of server timestamp
            long latency; // total latency from send to receive time
            long clockSkew; // clock skew between client and server
            long totalClockSkew=0; // sum of clock skew to calculate average
            long totalLatency=0; // sum of latency to calculate average
            long serverClockResolution=0; // clock resolution received from server
            String serverReply; // reply from the server
            String tokens[];

            for (int i=0;i<iterations+1;i++) {
                long sleepStart = System.nanoTime();
                Thread.sleep(50);
                clientSendNanos = System.nanoTime();

                //send & receive timestamp to server
                out.println(clientSendNanos);
                serverReply = in.readLine();

                clientReceiveNanos = System.nanoTime();

                tokens = serverReply.split(",");
                if (i==0) {
                    // first reply get clock resolution
                    serverClockResolution = (tokens.length>=2)?Long.parseLong(tokens[1]):clockResolution;
                    if (serverClockResolution>clockResolution) {
                        System.out.println("Server clock resolution ("+serverClockResolution+") applied because it is < client clock resolution ("+clockResolution+").");
                    } else {
                        System.out.println("Client clock resolution ("+serverClockResolution+") applied.");
                    }
                }else {
                    serverTimestamp = tokens[0];

                    serverNanos = Long.parseLong(serverTimestamp);
                    latency = clientReceiveNanos - clientSendNanos;
                    clockSkew = ((serverNanos - latency / 2) - clientSendNanos) / Math.max(clockResolution,serverClockResolution);
                    totalClockSkew+=clockSkew;
                    totalLatency+=latency;
                    System.out.println(i + " client send:" + clientSendNanos +
                            " server: " + serverTimestamp +
                            " client rcv: " + clientReceiveNanos +
                            " latency: " + latency +
                            " clock skew: " + clockSkew +
                            " sleep time: " + (clientSendNanos - sleepStart));
                }
            }
            System.out.println("Clock Resolution is "+clockResolution+" nanosecs.");
            System.out.println("Average Clock Skew is " + totalClockSkew/(long)iterations + " nanosecs.\n"+
                                // Following calculation rounds to nearest number and reduces to precision of clock
                               "Average Latency is " + (((totalLatency+5*clockResolution)/(long)iterations)/clockResolution)*clockResolution + " nanosecs rounded to clock resolution.");
        } catch (UnknownHostException e) {
            System.err.println("Unable to connect to " + hostName);
            System.exit(1);
        } catch (IOException e) {
            System.err.println("I/O Exception on Socket call to " +
                    hostName);
            System.exit(1);
        } catch (InterruptedException e) {
            System.err.println("Sleep faliure");
            System.exit(1);
        }
    }

    /**
     * Returns the current clock in nanoseconds
     *
     * @param portNumber - port number to accept upon
     */
    private static void server(int portNumber) {
        System.out.println("Attempting to accept connections on port "+portNumber+".");
        try (
                ServerSocket timeSocket = new ServerSocket(portNumber);
                //System.out.println("Waiting for connection...");
                Socket clientSocket = timeSocket.accept();
                PrintWriter out =
                        new PrintWriter(clientSocket.getOutputStream(), true);
                BufferedReader in = new BufferedReader(
                        new InputStreamReader(clientSocket.getInputStream()))
        ) {
            String inputLine;
            while ((inputLine = in.readLine()) != null) {
                System.out.println("Received "+inputLine);
                out.println(System.nanoTime()+","+clockResolution);
            }
        } catch (IOException e) {
            System.out.println("I/O Exception when trying to listen on port "
                    + portNumber);
            System.out.println(e.getMessage());
            System.exit(1);
        }
    }

    /**
     * Calls clockResolution to determine clock accuracy and then client or server depending on role.
     *
     * @param args - arguments to program
     */
    public static void main(String[] args) {
        String hostName = args[0];
        int portNumber = Integer.parseInt(args[1]);

        clockResolution = testClockResolution();

        if (hostName.equals("Server")) {
            server(portNumber);
        } else {
            int iterations = Integer.parseInt(args[2]);
            client(hostName, portNumber, iterations);
        }
    }
}